@@ -17,20 +17,14 @@
 #include <math.h>
 
 #include <php.h>
-#include <Zend/zend_interfaces.h>
+#include <zend_smart_str.h>
 #include <ext/date/php_date.h>
 #include <ext/standard/php_var.h>
-#include <zend_smart_str.h>
+#include <Zend/zend_interfaces.h>
 
-#ifdef HAVE_CONFIG_H
-#include "config.h"
-#endif
-
-#ifdef PHP_WIN32
-#include "win32/time.h"
-#endif
-#include "phongo_compat.h"
 #include "php_phongo.h"
+#include "phongo_error.h"
+#include "phongo_util.h"
 
 zend_class_entry* php_phongo_utcdatetime_ce;
 
@@ -82,7 +76,7 @@ static bool php_phongo_utcdatetime_init_from_current_time(php_phongo_utcdatetime
 	int64_t        sec, usec;
 	struct timeval cur_time;
 
-	gettimeofday(&cur_time, NULL);
+	bson_gettimeofday(&cur_time);
 	sec  = cur_time.tv_sec;
 	usec = cur_time.tv_usec;
 
@@ -99,12 +93,8 @@ static bool php_phongo_utcdatetime_init_from_date(php_phongo_utcdatetime_t* inte
 	int64_t sec, usec;
 
 	/* The following assignments use the same logic as date_format() in php_date.c */
-	sec = datetime_obj->time->sse;
-#if PHP_VERSION_ID >= 70200
+	sec  = datetime_obj->time->sse;
 	usec = (int64_t) floor(datetime_obj->time->us);
-#else
-	usec                  = (int64_t) floor(datetime_obj->time->f * 1000000 + 0.5);
-#endif
 
 	intern->milliseconds = (sec * 1000) + (usec / 1000);
 	intern->initialized  = true;
@@ -112,6 +102,29 @@ static bool php_phongo_utcdatetime_init_from_date(php_phongo_utcdatetime_t* inte
 	return true;
 } /* }}} */
 
+static HashTable* php_phongo_utcdatetime_get_properties_hash(phongo_compat_object_handler_type* object, bool is_temp) /* {{{ */
+{
+	php_phongo_utcdatetime_t* intern;
+	HashTable*                props;
+
+	intern = Z_OBJ_UTCDATETIME(PHONGO_COMPAT_GET_OBJ(object));
+
+	PHONGO_GET_PROPERTY_HASH_INIT_PROPS(is_temp, intern, props, 1);
+
+	if (!intern->initialized) {
+		return props;
+	}
+
+	{
+		zval milliseconds;
+
+		ZVAL_INT64_STRING(&milliseconds, intern->milliseconds);
+		zend_hash_str_update(props, "milliseconds", sizeof("milliseconds") - 1, &milliseconds);
+	}
+
+	return props;
+} /* }}} */
+
 /* {{{ proto void MongoDB\BSON\UTCDateTime::__construct([int|float|string|DateTimeInterface $milliseconds = null])
    Construct a new BSON UTCDateTime type from either the current time,
    milliseconds since the epoch, or a DateTimeInterface object. Defaults to the
@@ -239,11 +252,7 @@ static PHP_METHOD(UTCDateTime, toDateTime)
 	php_date_initialize(datetime_obj, sec, sec_len, NULL, NULL, 0);
 	efree(sec);
 
-#if PHP_VERSION_ID >= 70200
 	datetime_obj->time->us = (intern->milliseconds % 1000) * 1000;
-#else
-	datetime_obj->time->f = (double) (intern->milliseconds % 1000) / 1000;
-#endif
 }
 /* }}} */
 
@@ -341,7 +350,30 @@ static PHP_METHOD(UTCDateTime, unserialize)
 	zval_ptr_dtor(&props);
 } /* }}} */
 
+/* {{{ proto array MongoDB\Driver\UTCDateTime::__serialize()
+*/
+static PHP_METHOD(UTCDateTime, __serialize)
+{
+	PHONGO_PARSE_PARAMETERS_NONE();
+
+	RETURN_ARR(php_phongo_utcdatetime_get_properties_hash(PHONGO_COMPAT_OBJ_P(getThis()), true));
+} /* }}} */
+
+/* {{{ proto void MongoDB\Driver\UTCDateTime::__unserialize(array $data)
+*/
+static PHP_METHOD(UTCDateTime, __unserialize)
+{
+	zval* data;
+
+	PHONGO_PARSE_PARAMETERS_START(1, 1)
+	Z_PARAM_ARRAY(data)
+	PHONGO_PARSE_PARAMETERS_END();
+
+	php_phongo_utcdatetime_init_from_hash(Z_UTCDATETIME_OBJ_P(getThis()), Z_ARRVAL_P(data));
+} /* }}} */
+
 /* {{{ MongoDB\BSON\UTCDateTime function entries */
+/* clang-format off */
 ZEND_BEGIN_ARG_INFO_EX(ai_UTCDateTime___construct, 0, 0, 0)
 	ZEND_ARG_INFO(0, milliseconds)
 ZEND_END_ARG_INFO()
@@ -350,6 +382,16 @@ ZEND_BEGIN_ARG_INFO_EX(ai_UTCDateTime___set_state, 0, 0, 1)
 	ZEND_ARG_ARRAY_INFO(0, properties, 0)
 ZEND_END_ARG_INFO()
 
+ZEND_BEGIN_ARG_WITH_RETURN_TYPE_INFO_EX(ai_UTCDateTime___toString, 0, 0, IS_STRING, 0)
+ZEND_END_ARG_INFO()
+
+ZEND_BEGIN_ARG_INFO_EX(ai_UTCDateTime___unserialize, 0, 0, 1)
+	ZEND_ARG_ARRAY_INFO(0, data, 0)
+ZEND_END_ARG_INFO()
+
+ZEND_BEGIN_ARG_WITH_TENTATIVE_RETURN_TYPE_INFO_EX(ai_UTCDateTime_jsonSerialize, 0, 0, IS_ARRAY, 0)
+ZEND_END_ARG_INFO()
+
 ZEND_BEGIN_ARG_INFO_EX(ai_UTCDateTime_unserialize, 0, 0, 1)
 	ZEND_ARG_INFO(0, serialized)
 ZEND_END_ARG_INFO()
@@ -358,17 +400,18 @@ ZEND_BEGIN_ARG_INFO_EX(ai_UTCDateTime_void, 0, 0, 0)
 ZEND_END_ARG_INFO()
 
 static zend_function_entry php_phongo_utcdatetime_me[] = {
-	/* clang-format off */
 	PHP_ME(UTCDateTime, __construct, ai_UTCDateTime___construct, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
+	PHP_ME(UTCDateTime, __serialize, ai_UTCDateTime_void, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
 	PHP_ME(UTCDateTime, __set_state, ai_UTCDateTime___set_state, ZEND_ACC_PUBLIC | ZEND_ACC_STATIC)
-	PHP_ME(UTCDateTime, __toString, ai_UTCDateTime_void, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
-	PHP_ME(UTCDateTime, jsonSerialize, ai_UTCDateTime_void, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
+	PHP_ME(UTCDateTime, __toString, ai_UTCDateTime___toString, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
+	PHP_ME(UTCDateTime, __unserialize, ai_UTCDateTime___unserialize, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
+	PHP_ME(UTCDateTime, jsonSerialize, ai_UTCDateTime_jsonSerialize, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
 	PHP_ME(UTCDateTime, serialize, ai_UTCDateTime_void, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
 	PHP_ME(UTCDateTime, unserialize, ai_UTCDateTime_unserialize, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
 	PHP_ME(UTCDateTime, toDateTime, ai_UTCDateTime_void, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
 	PHP_FE_END
-	/* clang-format on */
 };
+/* clang-format on */
 /* }}} */
 
 /* {{{ MongoDB\BSON\UTCDateTime object handlers */
@@ -388,9 +431,7 @@ static void php_phongo_utcdatetime_free_object(zend_object* object) /* {{{ */
 
 static zend_object* php_phongo_utcdatetime_create_object(zend_class_entry* class_type) /* {{{ */
 {
-	php_phongo_utcdatetime_t* intern = NULL;
-
-	intern = PHONGO_ALLOC_OBJECT_T(php_phongo_utcdatetime_t, class_type);
+	php_phongo_utcdatetime_t* intern = zend_object_alloc(sizeof(php_phongo_utcdatetime_t), class_type);
 
 	zend_object_std_init(&intern->std, class_type);
 	object_properties_init(&intern->std, class_type);
@@ -433,29 +474,6 @@ static int php_phongo_utcdatetime_compare_objects(zval* o1, zval* o2) /* {{{ */
 	return 0;
 } /* }}} */
 
-static HashTable* php_phongo_utcdatetime_get_properties_hash(phongo_compat_object_handler_type* object, bool is_debug) /* {{{ */
-{
-	php_phongo_utcdatetime_t* intern;
-	HashTable*                props;
-
-	intern = Z_OBJ_UTCDATETIME(PHONGO_COMPAT_GET_OBJ(object));
-
-	PHONGO_GET_PROPERTY_HASH_INIT_PROPS(is_debug, intern, props, 1);
-
-	if (!intern->initialized) {
-		return props;
-	}
-
-	{
-		zval milliseconds;
-
-		ZVAL_INT64_STRING(&milliseconds, intern->milliseconds);
-		zend_hash_str_update(props, "milliseconds", sizeof("milliseconds") - 1, &milliseconds);
-	}
-
-	return props;
-} /* }}} */
-
 static HashTable* php_phongo_utcdatetime_get_debug_info(phongo_compat_object_handler_type* object, int* is_temp) /* {{{ */
 {
 	*is_temp = 1;
@@ -482,6 +500,10 @@ void php_phongo_utcdatetime_init_ce(INIT_FUNC_ARGS) /* {{{ */
 	zend_class_implements(php_phongo_utcdatetime_ce, 1, php_phongo_type_ce);
 	zend_class_implements(php_phongo_utcdatetime_ce, 1, zend_ce_serializable);
 
+#if PHP_VERSION_ID >= 80000
+	zend_class_implements(php_phongo_utcdatetime_ce, 1, zend_ce_stringable);
+#endif
+
 	memcpy(&php_phongo_handler_utcdatetime, phongo_get_std_object_handlers(), sizeof(zend_object_handlers));
 	PHONGO_COMPAT_SET_COMPARE_OBJECTS_HANDLER(utcdatetime);
 	php_phongo_handler_utcdatetime.clone_obj      = php_phongo_utcdatetime_clone_object;
@@ -490,12 +512,3 @@ void php_phongo_utcdatetime_init_ce(INIT_FUNC_ARGS) /* {{{ */
 	php_phongo_handler_utcdatetime.free_obj       = php_phongo_utcdatetime_free_object;
 	php_phongo_handler_utcdatetime.offset         = XtOffsetOf(php_phongo_utcdatetime_t, std);
 } /* }}} */
-
-/*
- * Local variables:
- * tab-width: 4
- * c-basic-offset: 4
- * End:
- * vim600: noet sw=4 ts=4 fdm=marker
- * vim<600: noet sw=4 ts=4
- */
