@@ -27,6 +27,8 @@
 #include "phongo_error.h"
 #include "phongo_util.h"
 
+#include "MongoDB/ClientEncryption.h"
+
 zend_class_entry* php_phongo_clientencryption_ce;
 
 /* Forward declarations */
@@ -34,6 +36,20 @@ static void phongo_clientencryption_create_datakey(php_phongo_clientencryption_t
 static void phongo_clientencryption_encrypt(php_phongo_clientencryption_t* clientencryption, zval* zvalue, zval* zciphertext, zval* options);
 static void phongo_clientencryption_decrypt(php_phongo_clientencryption_t* clientencryption, zval* zciphertext, zval* zvalue);
 
+/* {{{ proto void MongoDB\Driver\ClientEncryption::__construct(array $options)
+   Constructs a new ClientEncryption */
+static PHP_METHOD(ClientEncryption, __construct)
+{
+	zval* options;
+
+	PHONGO_PARSE_PARAMETERS_START(1, 1)
+	Z_PARAM_ARRAY(options)
+	PHONGO_PARSE_PARAMETERS_END();
+
+	/* An exception will be thrown on error. */
+	phongo_clientencryption_init(Z_CLIENTENCRYPTION_OBJ_P(getThis()), options, NULL);
+} /* }}} */
+
 /* {{{ proto MongoDB\BSON\Binary MongoDB\Driver\ClientEncryption::createDataKey(string $kmsProvider[, array $options])
    Creates a new key document and inserts into the key vault collection. */
 static PHP_METHOD(ClientEncryption, createDataKey)
@@ -97,6 +113,10 @@ static PHP_METHOD(ClientEncryption, decrypt)
 	phongo_clientencryption_decrypt(intern, ciphertext, return_value);
 } /* }}} */
 
+ZEND_BEGIN_ARG_INFO_EX(ai_ClientEncryption___construct, 0, 0, 0)
+	ZEND_ARG_ARRAY_INFO(0, options, 1)
+ZEND_END_ARG_INFO()
+
 ZEND_BEGIN_ARG_INFO_EX(ai_ClientEncryption_createDataKey, 0, 0, 1)
 	ZEND_ARG_INFO(0, kmsProvider)
 	ZEND_ARG_ARRAY_INFO(0, options, 1)
@@ -116,10 +136,10 @@ ZEND_END_ARG_INFO()
 
 static zend_function_entry php_phongo_clientencryption_me[] = {
 	/* clang-format off */
+	PHP_ME(ClientEncryption, __construct, ai_ClientEncryption___construct, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
 	PHP_ME(ClientEncryption, createDataKey, ai_ClientEncryption_createDataKey, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
 	PHP_ME(ClientEncryption, encrypt, ai_ClientEncryption_encrypt, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
 	PHP_ME(ClientEncryption, decrypt, ai_ClientEncryption_decrypt, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
-	ZEND_NAMED_ME(__construct, PHP_FN(MongoDB_disabled___construct), ai_ClientEncryption_void, ZEND_ACC_PRIVATE | ZEND_ACC_FINAL)
 	ZEND_NAMED_ME(__wakeup, PHP_FN(MongoDB_disabled___wakeup), ai_ClientEncryption_void, ZEND_ACC_PUBLIC | ZEND_ACC_FINAL)
 	PHP_FE_END
 	/* clang-format on */
@@ -189,17 +209,19 @@ void php_phongo_clientencryption_init_ce(INIT_FUNC_ARGS) /* {{{ */
 
 	zend_declare_class_constant_string(php_phongo_clientencryption_ce, ZEND_STRL("AEAD_AES_256_CBC_HMAC_SHA_512_DETERMINISTIC"), MONGOC_AEAD_AES_256_CBC_HMAC_SHA_512_DETERMINISTIC);
 	zend_declare_class_constant_string(php_phongo_clientencryption_ce, ZEND_STRL("AEAD_AES_256_CBC_HMAC_SHA_512_RANDOM"), MONGOC_AEAD_AES_256_CBC_HMAC_SHA_512_RANDOM);
+	zend_declare_class_constant_string(php_phongo_clientencryption_ce, ZEND_STRL("ALGORITHM_INDEXED"), MONGOC_ENCRYPT_ALGORITHM_INDEXED);
+	zend_declare_class_constant_string(php_phongo_clientencryption_ce, ZEND_STRL("ALGORITHM_UNINDEXED"), MONGOC_ENCRYPT_ALGORITHM_UNINDEXED);
+	zend_declare_class_constant_string(php_phongo_clientencryption_ce, ZEND_STRL("QUERY_TYPE_EQUALITY"), MONGOC_ENCRYPT_QUERY_TYPE_EQUALITY);
 } /* }}} */
 
 #ifdef MONGOC_ENABLE_CLIENT_SIDE_ENCRYPTION
-/* keyVaultClientManager is an output parameter and will be assigned the
- * keyVaultNamespace Manager (if any). */
-static mongoc_client_encryption_opts_t* phongo_clientencryption_opts_from_zval(zval* defaultKeyVaultClient, zval* options, zval** keyVaultClientManager) /* {{{ */
+/* key_vault_client_manager is an output parameter and will be assigned to the
+ * effective keyVaultClient. */
+static mongoc_client_encryption_opts_t* phongo_clientencryption_opts_from_zval(zval* options, zval* default_key_vault_client_manager, zval** key_vault_client_manager) /* {{{ */
 {
-	mongoc_client_encryption_opts_t* opts;
+	mongoc_client_encryption_opts_t* opts = mongoc_client_encryption_opts_new();
 
-	opts                   = mongoc_client_encryption_opts_new();
-	*keyVaultClientManager = NULL;
+	*key_vault_client_manager = NULL;
 
 	if (!options || Z_TYPE_P(options) != IS_ARRAY) {
 		/* Returning opts as-is will defer to mongoc_client_encryption_new to
@@ -211,15 +233,20 @@ static mongoc_client_encryption_opts_t* phongo_clientencryption_opts_from_zval(z
 		zval* key_vault_client = php_array_fetch(options, "keyVaultClient");
 
 		if (Z_TYPE_P(key_vault_client) != IS_OBJECT || !instanceof_function(Z_OBJCE_P(key_vault_client), php_phongo_manager_ce)) {
-			phongo_throw_exception(PHONGO_ERROR_INVALID_ARGUMENT, "Expected \"keyVaultClient\" encryption option to be %s, %s given", ZSTR_VAL(php_phongo_manager_ce->name), PHONGO_ZVAL_CLASS_OR_TYPE_NAME_P(key_vault_client));
+			phongo_throw_exception(PHONGO_ERROR_INVALID_ARGUMENT, "Expected \"keyVaultClient\" option to be %s, %s given", ZSTR_VAL(php_phongo_manager_ce->name), PHONGO_ZVAL_CLASS_OR_TYPE_NAME_P(key_vault_client));
 			goto cleanup;
 		}
 
 		mongoc_client_encryption_opts_set_keyvault_client(opts, Z_MANAGER_OBJ_P(key_vault_client)->client);
-		*keyVaultClientManager = key_vault_client;
+		*key_vault_client_manager = key_vault_client;
+	} else if (default_key_vault_client_manager) {
+		mongoc_client_encryption_opts_set_keyvault_client(opts, Z_MANAGER_OBJ_P(default_key_vault_client_manager)->client);
+		*key_vault_client_manager = default_key_vault_client_manager;
 	} else {
-		mongoc_client_encryption_opts_set_keyvault_client(opts, Z_MANAGER_OBJ_P(defaultKeyVaultClient)->client);
-		*keyVaultClientManager = defaultKeyVaultClient;
+		/* If the ClientEncryption object is being constructed directly, the
+		 * "keyVaultClient" option must be explicitly provided. */
+		phongo_throw_exception(PHONGO_ERROR_INVALID_ARGUMENT, "The \"keyVaultClient\" option is required when constructing a ClientEncryption object directly");
+		goto cleanup;
 	}
 
 	if (php_array_existsc(options, "keyVaultNamespace")) {
@@ -232,7 +259,7 @@ static mongoc_client_encryption_opts_t* phongo_clientencryption_opts_from_zval(z
 		keyvault_namespace = php_array_fetchc_string(options, "keyVaultNamespace", &plen, &pfree);
 
 		if (!phongo_split_namespace(keyvault_namespace, &db_name, &coll_name)) {
-			phongo_throw_exception(PHONGO_ERROR_INVALID_ARGUMENT, "Expected \"keyVaultNamespace\" encryption option to contain a full collection name");
+			phongo_throw_exception(PHONGO_ERROR_INVALID_ARGUMENT, "Expected \"keyVaultNamespace\" option to contain a full collection namespace");
 
 			if (pfree) {
 				efree(keyvault_namespace);
@@ -255,7 +282,7 @@ static mongoc_client_encryption_opts_t* phongo_clientencryption_opts_from_zval(z
 		bson_t bson_providers = BSON_INITIALIZER;
 
 		if (Z_TYPE_P(kms_providers) != IS_ARRAY && Z_TYPE_P(kms_providers) != IS_OBJECT) {
-			phongo_throw_exception(PHONGO_ERROR_INVALID_ARGUMENT, "Expected \"kmsProviders\" encryption option to be an array or object");
+			phongo_throw_exception(PHONGO_ERROR_INVALID_ARGUMENT, "Expected \"kmsProviders\" option to be an array or object, %s given", PHONGO_ZVAL_CLASS_OR_TYPE_NAME_P(kms_providers));
 			goto cleanup;
 		}
 
@@ -273,7 +300,7 @@ static mongoc_client_encryption_opts_t* phongo_clientencryption_opts_from_zval(z
 		bson_t bson_options = BSON_INITIALIZER;
 
 		if (Z_TYPE_P(tls_options) != IS_ARRAY && Z_TYPE_P(tls_options) != IS_OBJECT) {
-			phongo_throw_exception(PHONGO_ERROR_INVALID_ARGUMENT, "Expected \"tlsOptions\" encryption option to be an array or object");
+			phongo_throw_exception(PHONGO_ERROR_INVALID_ARGUMENT, "Expected \"tlsOptions\" option to be an array or object, %s given", PHONGO_ZVAL_CLASS_OR_TYPE_NAME_P(tls_options));
 			goto cleanup;
 		}
 
@@ -296,32 +323,35 @@ cleanup:
 	return NULL;
 } /* }}} */
 
-void phongo_clientencryption_init(zval* return_value, zval* manager, zval* options) /* {{{ */
+void phongo_clientencryption_init(php_phongo_clientencryption_t* intern, zval* options, zval* default_key_vault_client_manager) /* {{{ */
 {
-	php_phongo_clientencryption_t*   intern;
 	mongoc_client_encryption_t*      client_encryption;
 	mongoc_client_encryption_opts_t* opts;
-	zval*                            key_vault_client_manager = manager;
+	zval*                            key_vault_client_manager = NULL;
 	bson_error_t                     error                    = { 0 };
 
-	opts = phongo_clientencryption_opts_from_zval(manager, options, &key_vault_client_manager);
+	opts = phongo_clientencryption_opts_from_zval(options, default_key_vault_client_manager, &key_vault_client_manager);
+
 	if (!opts) {
 		/* Exception already thrown */
 		goto cleanup;
 	}
 
 	client_encryption = mongoc_client_encryption_new(opts, &error);
+
 	if (!client_encryption) {
 		phongo_throw_exception_from_bson_error_t(&error);
-
 		goto cleanup;
 	}
 
-	object_init_ex(return_value, php_phongo_clientencryption_ce);
-
-	intern                    = Z_CLIENTENCRYPTION_OBJ_P(return_value);
 	intern->client_encryption = client_encryption;
-	ZVAL_ZVAL(&intern->key_vault_client_manager, key_vault_client_manager, 1, 0);
+
+	/* Note: key_vault_client_manager should always be assigned if options were
+	 * successfully parsed by phongo_clientencryption_opts_from_zval, but let's
+	 * be defensive. */
+	if (key_vault_client_manager) {
+		ZVAL_ZVAL(&intern->key_vault_client_manager, key_vault_client_manager, 1, 0);
+	}
 
 cleanup:
 	if (opts) {
@@ -425,10 +455,11 @@ cleanup:
 static void phongo_clientencryption_create_datakey(php_phongo_clientencryption_t* clientencryption, zval* return_value, char* kms_provider, zval* options) /* {{{ */
 {
 	mongoc_client_encryption_datakey_opts_t* opts;
-	bson_value_t                             keyid;
+	bson_value_t                             keyid = { 0 };
 	bson_error_t                             error = { 0 };
 
 	opts = phongo_clientencryption_datakey_opts_from_zval(options);
+
 	if (!opts) {
 		/* Exception already thrown */
 		goto cleanup;
@@ -448,6 +479,8 @@ cleanup:
 	if (opts) {
 		mongoc_client_encryption_datakey_opts_destroy(opts);
 	}
+
+	bson_value_destroy(&keyid);
 } /* }}} */
 
 static mongoc_client_encryption_encrypt_opts_t* phongo_clientencryption_encrypt_opts_from_zval(zval* options) /* {{{ */
@@ -460,15 +493,22 @@ static mongoc_client_encryption_encrypt_opts_t* phongo_clientencryption_encrypt_
 		return opts;
 	}
 
+	if (php_array_existsc(options, "contentionFactor")) {
+		mongoc_client_encryption_encrypt_opts_set_contention_factor(opts, php_array_fetch_long(options, "contentionFactor"));
+	}
+
 	if (php_array_existsc(options, "keyId")) {
-		bson_value_t keyid;
+		bson_value_t keyid = { 0 };
 
 		php_phongo_zval_to_bson_value(php_array_fetchc(options, "keyId"), PHONGO_BSON_NONE, &keyid);
+
 		if (EG(exception)) {
+			bson_value_destroy(&keyid);
 			goto cleanup;
 		}
 
 		mongoc_client_encryption_encrypt_opts_set_keyid(opts, &keyid);
+		bson_value_destroy(&keyid);
 	}
 
 	if (php_array_existsc(options, "keyAltName")) {
@@ -497,6 +537,19 @@ static mongoc_client_encryption_encrypt_opts_t* phongo_clientencryption_encrypt_
 		}
 	}
 
+	if (php_array_existsc(options, "queryType")) {
+		char*     querytype;
+		int       plen;
+		zend_bool pfree;
+
+		querytype = php_array_fetch_string(options, "queryType", &plen, &pfree);
+		mongoc_client_encryption_encrypt_opts_set_query_type(opts, querytype);
+
+		if (pfree) {
+			efree(querytype);
+		}
+	}
+
 	return opts;
 
 cleanup:
@@ -510,12 +563,14 @@ cleanup:
 static void phongo_clientencryption_encrypt(php_phongo_clientencryption_t* clientencryption, zval* zvalue, zval* zciphertext, zval* options) /* {{{ */
 {
 	mongoc_client_encryption_encrypt_opts_t* opts;
-	bson_value_t                             ciphertext, value;
-	bson_error_t                             error = { 0 };
+	bson_value_t                             ciphertext = { 0 };
+	bson_value_t                             value      = { 0 };
+	bson_error_t                             error      = { 0 };
 
 	php_phongo_zval_to_bson_value(zvalue, PHONGO_BSON_NONE, &value);
 
 	opts = phongo_clientencryption_encrypt_opts_from_zval(options);
+
 	if (!opts) {
 		/* Exception already thrown */
 		goto cleanup;
@@ -535,27 +590,35 @@ cleanup:
 	if (opts) {
 		mongoc_client_encryption_encrypt_opts_destroy(opts);
 	}
+
+	bson_value_destroy(&ciphertext);
+	bson_value_destroy(&value);
 } /* }}} */
 
 static void phongo_clientencryption_decrypt(php_phongo_clientencryption_t* clientencryption, zval* zciphertext, zval* zvalue) /* {{{ */
 {
-	bson_value_t ciphertext, value;
-	bson_error_t error = { 0 };
+	bson_value_t ciphertext = { 0 };
+	bson_value_t value      = { 0 };
+	bson_error_t error      = { 0 };
 
 	php_phongo_zval_to_bson_value(zciphertext, PHONGO_BSON_NONE, &ciphertext);
 
 	if (!mongoc_client_encryption_decrypt(clientencryption->client_encryption, &ciphertext, &value, &error)) {
 		phongo_throw_exception_from_bson_error_t(&error);
-		return;
+		goto cleanup;
 	}
 
 	if (!php_phongo_bson_value_to_zval(&value, zvalue)) {
 		/* Exception already thrown */
-		return;
+		goto cleanup;
 	}
+
+cleanup:
+	bson_value_destroy(&ciphertext);
+	bson_value_destroy(&value);
 } /* }}} */
 #else  /* MONGOC_ENABLE_CLIENT_SIDE_ENCRYPTION */
-void phongo_clientencryption_init(php_phongo_clientencryption_t* clientencryption, zval* manager, zval* options) /* {{{ */
+void phongo_clientencryption_init(php_phongo_clientencryption_t* intern, zval* options, zval* default_key_vault_client_manager) /* {{{ */
 {
 	phongo_throw_exception_no_cse(PHONGO_ERROR_RUNTIME, "Cannot configure clientEncryption object.");
 }
