@@ -1,5 +1,5 @@
 /*
- * Copyright 2014-2017 MongoDB, Inc.
+ * Copyright 2014-present MongoDB, Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -14,17 +14,19 @@
  * limitations under the License.
  */
 
+#include "bson/bson.h"
+#include "mongoc/mongoc.h"
+
 #include <php.h>
 #include <Zend/zend_interfaces.h>
 
-#ifdef HAVE_CONFIG_H
-#include "config.h"
-#endif
-
 #include "php_array_api.h"
-#include "phongo_compat.h"
+
 #include "php_phongo.h"
-#include "php_bson.h"
+#include "phongo_bson_encode.h"
+#include "phongo_error.h"
+
+#include "MongoDB/ReadConcern.h"
 
 zend_class_entry* php_phongo_query_ce;
 
@@ -82,6 +84,30 @@ static bool php_phongo_query_opts_append_document(bson_t* opts, const char* opts
 	return true;
 } /* }}} */
 
+/* Appends an arbitrary BSON value for the given opts document and key. Returns
+ * true on success; otherwise, false is returned and an exception is thrown. */
+static bool php_phongo_query_opts_append_value(bson_t* opts, const char* opts_key, zval* zarr, const char* zarr_key) /* {{{ */
+{
+	bson_value_t value = { 0 };
+
+	php_phongo_zval_to_bson_value(php_array_fetch(zarr, zarr_key), PHONGO_BSON_NONE, &value);
+
+	if (EG(exception)) {
+		/* Exception should already have been thrown */
+		bson_value_destroy(&value);
+		return false;
+	}
+
+	if (!BSON_APPEND_VALUE(opts, opts_key, &value)) {
+		phongo_throw_exception(PHONGO_ERROR_INVALID_ARGUMENT, "Error appending \"%s\" option", opts_key);
+		bson_value_destroy(&value);
+		return false;
+	}
+
+	bson_value_destroy(&value);
+	return true;
+} /* }}} */
+
 #define PHONGO_QUERY_OPT_BOOL_EX(opt, zarr, key, deprecated)                                                                      \
 	if ((zarr) && php_array_existsc((zarr), (key))) {                                                                             \
 		if ((deprecated)) {                                                                                                       \
@@ -96,6 +122,13 @@ static bool php_phongo_query_opts_append_document(bson_t* opts, const char* opts
 #define PHONGO_QUERY_OPT_BOOL(opt, zarr, key) PHONGO_QUERY_OPT_BOOL_EX((opt), (zarr), (key), 0)
 #define PHONGO_QUERY_OPT_BOOL_DEPRECATED(opt, zarr, key) PHONGO_QUERY_OPT_BOOL_EX((opt), (zarr), (key), 1)
 
+#define PHONGO_QUERY_OPT_BSON_VALUE(opt, zarr, key)                                    \
+	if ((zarr) && php_array_existsc((zarr), (key))) {                                  \
+		if (!php_phongo_query_opts_append_value(intern->opts, (opt), (zarr), (key))) { \
+			return false;                                                              \
+		}                                                                              \
+	}
+
 #define PHONGO_QUERY_OPT_DOCUMENT(opt, zarr, key)                                         \
 	if ((zarr) && php_array_existsc((zarr), (key))) {                                     \
 		if (!php_phongo_query_opts_append_document(intern->opts, (opt), (zarr), (key))) { \
@@ -294,9 +327,10 @@ static bool php_phongo_query_init(php_phongo_query_t* intern, zval* filter, zval
 	PHONGO_QUERY_OPT_BOOL("awaitData", options, "awaitData");
 	PHONGO_QUERY_OPT_INT64("batchSize", options, "batchSize");
 	PHONGO_QUERY_OPT_DOCUMENT("collation", options, "collation");
-	PHONGO_QUERY_OPT_STRING("comment", options, "comment")
-	else PHONGO_QUERY_OPT_STRING("comment", modifiers, "$comment");
+	PHONGO_QUERY_OPT_BSON_VALUE("comment", options, "comment")
+	else PHONGO_QUERY_OPT_BSON_VALUE("comment", modifiers, "$comment");
 	PHONGO_QUERY_OPT_BOOL("exhaust", options, "exhaust");
+	PHONGO_QUERY_OPT_DOCUMENT("let", options, "let");
 	PHONGO_QUERY_OPT_DOCUMENT("max", options, "max")
 	else PHONGO_QUERY_OPT_DOCUMENT("max", modifiers, "$max");
 	PHONGO_QUERY_OPT_INT64_DEPRECATED("maxScan", options, "maxScan")
@@ -343,23 +377,28 @@ static bool php_phongo_query_init(php_phongo_query_t* intern, zval* filter, zval
 	return true;
 } /* }}} */
 
+#undef PHONGO_QUERY_OPT_BOOL_EX
 #undef PHONGO_QUERY_OPT_BOOL
+#undef PHONGO_QUERY_OPT_BOOL_DEPRECATED
+#undef PHONGO_QUERY_OPT_BSON_VALUE
 #undef PHONGO_QUERY_OPT_DOCUMENT
+#undef PHONGO_QUERY_OPT_INT64_EX
 #undef PHONGO_QUERY_OPT_INT64
+#undef PHONGO_QUERY_OPT_INT64_DEPRECATED
 #undef PHONGO_QUERY_OPT_STRING
 
 /* {{{ proto void MongoDB\Driver\Query::__construct(array|object $filter[, array $options = array()])
    Constructs a new Query */
 static PHP_METHOD(Query, __construct)
 {
-	php_phongo_query_t* intern;
 	zend_error_handling error_handling;
+	php_phongo_query_t* intern;
 	zval*               filter;
 	zval*               options = NULL;
 
-	zend_replace_error_handling(EH_THROW, phongo_exception_from_phongo_domain(PHONGO_ERROR_INVALID_ARGUMENT), &error_handling);
 	intern = Z_QUERY_OBJ_P(getThis());
 
+	zend_replace_error_handling(EH_THROW, phongo_exception_from_phongo_domain(PHONGO_ERROR_INVALID_ARGUMENT), &error_handling);
 	if (zend_parse_parameters(ZEND_NUM_ARGS(), "A|a!", &filter, &options) == FAILURE) {
 		zend_restore_error_handling(&error_handling);
 		return;
@@ -411,9 +450,7 @@ static void php_phongo_query_free_object(zend_object* object) /* {{{ */
 
 static zend_object* php_phongo_query_create_object(zend_class_entry* class_type) /* {{{ */
 {
-	php_phongo_query_t* intern = NULL;
-
-	intern = PHONGO_ALLOC_OBJECT_T(php_phongo_query_t, class_type);
+	php_phongo_query_t* intern = zend_object_alloc(sizeof(php_phongo_query_t), class_type);
 
 	zend_object_std_init(&intern->std, class_type);
 	object_properties_init(&intern->std, class_type);
@@ -423,13 +460,13 @@ static zend_object* php_phongo_query_create_object(zend_class_entry* class_type)
 	return &intern->std;
 } /* }}} */
 
-static HashTable* php_phongo_query_get_debug_info(zval* object, int* is_temp) /* {{{ */
+static HashTable* php_phongo_query_get_debug_info(phongo_compat_object_handler_type* object, int* is_temp) /* {{{ */
 {
 	php_phongo_query_t* intern;
 	zval                retval = ZVAL_STATIC_INIT;
 
 	*is_temp = 1;
-	intern   = Z_QUERY_OBJ_P(object);
+	intern   = Z_OBJ_QUERY(PHONGO_COMPAT_GET_OBJ(object));
 
 	array_init_size(&retval, 3);
 
@@ -491,12 +528,3 @@ void php_phongo_query_init_ce(INIT_FUNC_ARGS) /* {{{ */
 	php_phongo_handler_query.free_obj       = php_phongo_query_free_object;
 	php_phongo_handler_query.offset         = XtOffsetOf(php_phongo_query_t, std);
 } /* }}} */
-
-/*
- * Local variables:
- * tab-width: 4
- * c-basic-offset: 4
- * End:
- * vim600: noet sw=4 ts=4 fdm=marker
- * vim<600: noet sw=4 ts=4
- */
