@@ -103,13 +103,13 @@ _mongoc_client_killcursors_command (mongoc_cluster_t *cluster,
 
 #ifdef MONGOC_HAVE_DNSAPI
 
-typedef bool (*mongoc_rr_callback_t) (const char *service,
+typedef bool (*mongoc_rr_callback_t) (const char *hostname,
                                       PDNS_RECORD pdns,
                                       mongoc_rr_data_t *rr_data,
                                       bson_error_t *error);
 
 static bool
-srv_callback (const char *service,
+srv_callback (const char *hostname,
               PDNS_RECORD pdns,
               mongoc_rr_data_t *rr_data,
               bson_error_t *error)
@@ -132,7 +132,7 @@ srv_callback (const char *service,
 
 /* rr_data is unused, but here to match srv_callback signature */
 static bool
-txt_callback (const char *service,
+txt_callback (const char *hostname,
               PDNS_RECORD pdns,
               mongoc_rr_data_t *rr_data,
               bson_error_t *error)
@@ -178,7 +178,7 @@ txt_callback (const char *service,
  */
 
 static bool
-_mongoc_get_rr_dnsapi (const char *service,
+_mongoc_get_rr_dnsapi (const char *hostname,
                        mongoc_rr_type_t rr_type,
                        mongoc_rr_data_t *rr_data,
                        bson_error_t *error)
@@ -209,7 +209,7 @@ _mongoc_get_rr_dnsapi (const char *service,
       callback = txt_callback;
    }
 
-   res = DnsQuery_UTF8 (service,
+   res = DnsQuery_UTF8 (hostname,
                         nst,
                         DNS_QUERY_BYPASS_CACHE,
                         NULL /* IP Address */,
@@ -229,17 +229,17 @@ _mongoc_get_rr_dnsapi (const char *service,
                          0)) {
          DNS_ERROR ("Failed to look up %s record \"%s\": %s",
                     rr_type_name,
-                    service,
+                    hostname,
                     (char *) lpMsgBuf);
       }
 
       DNS_ERROR ("Failed to look up %s record \"%s\": Unknown error",
                  rr_type_name,
-                 service);
+                 hostname);
    }
 
    if (!pdns) {
-      DNS_ERROR ("No %s records for \"%s\"", rr_type_name, service);
+      DNS_ERROR ("No %s records for \"%s\"", rr_type_name, hostname);
    }
 
    i = 0;
@@ -252,7 +252,7 @@ _mongoc_get_rr_dnsapi (const char *service,
             /* Initial DNS Seedlist Discovery Spec: a client "MUST raise an
             error when multiple TXT records are encountered". */
             callback_success = false;
-            DNS_ERROR ("Multiple TXT records for \"%s\"", service);
+            DNS_ERROR ("Multiple TXT records for \"%s\"", hostname);
          }
 
          if (rr_data) {
@@ -261,7 +261,7 @@ _mongoc_get_rr_dnsapi (const char *service,
             }
          }
 
-         if (!callback (service, pdns, rr_data, error)) {
+         if (!callback (hostname, pdns, rr_data, error)) {
             callback_success = false;
             GOTO (done);
          }
@@ -275,7 +275,7 @@ _mongoc_get_rr_dnsapi (const char *service,
 
    rr_data->count = i;
    if (i == 0) {
-      DNS_ERROR ("No matching %s records for \"%s\"", rr_type_name, service);
+      DNS_ERROR ("No matching %s records for \"%s\"", rr_type_name, hostname);
    }
    dns_success = true;
 
@@ -293,14 +293,32 @@ done:
 
 #elif (defined(MONGOC_HAVE_RES_NSEARCH) || defined(MONGOC_HAVE_RES_SEARCH))
 
-typedef bool (*mongoc_rr_callback_t) (const char *service,
+typedef bool (*mongoc_rr_callback_t) (const char *hostname,
                                       ns_msg *ns_answer,
                                       ns_rr *rr,
                                       mongoc_rr_data_t *rr_data,
                                       bson_error_t *error);
 
+static const char *
+_mongoc_hstrerror (int code)
+{
+   switch (code) {
+   case HOST_NOT_FOUND:
+      return "The specified host is unknown.";
+   case NO_ADDRESS:
+      return "The requested name is valid but does not have an IP address.";
+   case NO_RECOVERY:
+      return "A nonrecoverable name server error occurred.";
+   case TRY_AGAIN:
+      return "A temporary error occurred on an authoritative name server. Try "
+             "again later.";
+   default:
+      return "An unknown error occurred.";
+   }
+}
+
 static bool
-srv_callback (const char *service,
+srv_callback (const char *hostname,
               ns_msg *ns_answer,
               ns_rr *rr,
               mongoc_rr_data_t *rr_data,
@@ -327,8 +345,8 @@ srv_callback (const char *service,
 
    if (size < 1) {
       DNS_ERROR ("Invalid record in SRV answer for \"%s\": \"%s\"",
-                 service,
-                 strerror (h_errno));
+                 hostname,
+                 _mongoc_hstrerror (h_errno));
    }
 
    if (!_mongoc_host_list_from_hostport_with_err (
@@ -342,7 +360,7 @@ done:
 }
 
 static bool
-txt_callback (const char *service,
+txt_callback (const char *hostname,
               ns_msg *ns_answer,
               ns_rr *rr,
               mongoc_rr_data_t *rr_data,
@@ -357,7 +375,7 @@ txt_callback (const char *service,
 
    total = (uint16_t) ns_rr_rdlen (*rr);
    if (total < 1 || total > 255) {
-      DNS_ERROR ("Invalid TXT record size %hu for \"%s\"", total, service);
+      DNS_ERROR ("Invalid TXT record size %hu for \"%s\"", total, hostname);
    }
 
    /* a TXT record has one or more strings, each up to 255 chars, each is
@@ -408,9 +426,10 @@ done:
  */
 
 static bool
-_mongoc_get_rr_search (const char *service,
+_mongoc_get_rr_search (const char *hostname,
                        mongoc_rr_type_t rr_type,
                        mongoc_rr_data_t *rr_data,
+                       size_t initial_buffer_size,
                        bson_error_t *error)
 {
 #ifdef MONGOC_HAVE_RES_NSEARCH
@@ -418,7 +437,7 @@ _mongoc_get_rr_search (const char *service,
 #endif
    int size = 0;
    unsigned char *search_buf = NULL;
-   size_t buffer_size = 1024;
+   size_t buffer_size = initial_buffer_size;
    ns_msg ns_answer;
    int n;
    int i;
@@ -429,6 +448,7 @@ _mongoc_get_rr_search (const char *service,
    bool dns_success;
    bool callback_success = true;
    int num_matching_records;
+   uint32_t ttl;
 
    ENTRY;
 
@@ -463,26 +483,26 @@ _mongoc_get_rr_search (const char *service,
       /* thread-safe */
       res_ninit (&state);
       size =
-         res_nsearch (&state, service, ns_c_in, nst, search_buf, buffer_size);
+         res_nsearch (&state, hostname, ns_c_in, nst, search_buf, buffer_size);
 #elif defined(MONGOC_HAVE_RES_SEARCH)
-      size = res_search (service, ns_c_in, nst, search_buf, buffer_size);
+      size = res_search (hostname, ns_c_in, nst, search_buf, buffer_size);
 #endif
 
       if (size < 0) {
          DNS_ERROR ("Failed to look up %s record \"%s\": %s",
                     rr_type_name,
-                    service,
-                    strerror (h_errno));
+                    hostname,
+                    _mongoc_hstrerror (h_errno));
       }
-   } while (size > buffer_size);
+   } while (size >= buffer_size);
 
    if (ns_initparse (search_buf, size, &ns_answer)) {
-      DNS_ERROR ("Invalid %s answer for \"%s\"", rr_type_name, service);
+      DNS_ERROR ("Invalid %s answer for \"%s\"", rr_type_name, hostname);
    }
 
    n = ns_msg_count (ns_answer, ns_s_an);
    if (!n) {
-      DNS_ERROR ("No %s records for \"%s\"", rr_type_name, service);
+      DNS_ERROR ("No %s records for \"%s\"", rr_type_name, hostname);
    }
 
    rr_data->count = n;
@@ -492,8 +512,8 @@ _mongoc_get_rr_search (const char *service,
          DNS_ERROR ("Invalid record %d of %s answer for \"%s\": \"%s\"",
                     i,
                     rr_type_name,
-                    service,
-                    strerror (h_errno));
+                    hostname,
+                    _mongoc_hstrerror (h_errno));
       }
 
       /* Skip records that don't match the ones we requested. CDRIVER-3628 shows
@@ -512,28 +532,24 @@ _mongoc_get_rr_search (const char *service,
          /* Initial DNS Seedlist Discovery Spec: a client "MUST raise an error
           * when multiple TXT records are encountered". */
          callback_success = false;
-         DNS_ERROR ("Multiple TXT records for \"%s\"", service);
+         DNS_ERROR ("Multiple TXT records for \"%s\"", hostname);
       }
 
       num_matching_records++;
 
-      if (rr_data) {
-         uint32_t ttl;
-
-         ttl = ns_rr_ttl (resource_record);
-         if ((i == 0) || (ttl < rr_data->min_ttl)) {
-            rr_data->min_ttl = ttl;
-         }
+      ttl = ns_rr_ttl (resource_record);
+      if ((i == 0) || (ttl < rr_data->min_ttl)) {
+         rr_data->min_ttl = ttl;
       }
 
-      if (!callback (service, &ns_answer, &resource_record, rr_data, error)) {
+      if (!callback (hostname, &ns_answer, &resource_record, rr_data, error)) {
          callback_success = false;
          GOTO (done);
       }
    }
 
    if (num_matching_records == 0) {
-      DNS_ERROR ("No matching %s records for \"%s\"", rr_type_name, service);
+      DNS_ERROR ("No matching %s records for \"%s\"", rr_type_name, hostname);
    }
 
    dns_success = true;
@@ -579,15 +595,19 @@ done:
  */
 
 bool
-_mongoc_client_get_rr (const char *service,
+_mongoc_client_get_rr (const char *hostname,
                        mongoc_rr_type_t rr_type,
                        mongoc_rr_data_t *rr_data,
+                       size_t initial_buffer_size,
                        bson_error_t *error)
 {
+   BSON_ASSERT (rr_data);
+
 #ifdef MONGOC_HAVE_DNSAPI
-   return _mongoc_get_rr_dnsapi (service, rr_type, rr_data, error);
+   return _mongoc_get_rr_dnsapi (hostname, rr_type, rr_data, error);
 #elif (defined(MONGOC_HAVE_RES_NSEARCH) || defined(MONGOC_HAVE_RES_SEARCH))
-   return _mongoc_get_rr_search (service, rr_type, rr_data, error);
+   return _mongoc_get_rr_search (
+      hostname, rr_type, rr_data, initial_buffer_size, error);
 #else
    bson_set_error (error,
                    MONGOC_ERROR_STREAM,
@@ -962,48 +982,27 @@ _mongoc_client_recv (mongoc_client_t *client,
 }
 
 
-/*
- *--------------------------------------------------------------------------
- *
- * mongoc_client_new --
- *
- *       Create a new mongoc_client_t using the URI provided.
- *
- *       @uri should be a MongoDB URI string such as "mongodb://localhost/"
- *       More information on the format can be found at
- *       http://docs.mongodb.org/manual/reference/connection-string/
- *
- * Returns:
- *       A newly allocated mongoc_client_t or NULL if @uri_string is
- *       invalid.
- *
- * Side effects:
- *       None.
- *
- *--------------------------------------------------------------------------
- */
 mongoc_client_t *
 mongoc_client_new (const char *uri_string)
 {
-   mongoc_topology_t *topology;
    mongoc_client_t *client;
    mongoc_uri_t *uri;
-
+   bson_error_t error = {0};
 
    if (!uri_string) {
       uri_string = "mongodb://127.0.0.1/";
    }
 
-   if (!(uri = mongoc_uri_new (uri_string))) {
+   if (!(uri = mongoc_uri_new_with_error (uri_string, &error))) {
+      /* Log URI errors as a warning for consistency with mongoc_uri_new */
+      MONGOC_WARNING ("Error parsing URI: '%s'", error.message);
       return NULL;
    }
 
-   topology = mongoc_topology_new (uri, true);
-
-   client = _mongoc_client_new_from_uri (topology);
-   if (!client) {
-      mongoc_topology_destroy (topology);
+   if (!(client = mongoc_client_new_from_uri_with_error (uri, &error))) {
+      MONGOC_ERROR ("%s", error.message);
    }
+
    mongoc_uri_destroy (uri);
 
    return client;
@@ -1064,54 +1063,65 @@ mongoc_client_set_ssl_opts (mongoc_client_t *client,
 #endif
 
 
-/*
- *--------------------------------------------------------------------------
- *
- * mongoc_client_new_from_uri --
- *
- *       Create a new mongoc_client_t for a mongoc_uri_t.
- *
- * Returns:
- *       A newly allocated mongoc_client_t.
- *
- * Side effects:
- *       None.
- *
- *--------------------------------------------------------------------------
- */
-
 mongoc_client_t *
 mongoc_client_new_from_uri (const mongoc_uri_t *uri)
 {
+   mongoc_client_t *client;
+   bson_error_t error = {0};
+
+   if (!(client = mongoc_client_new_from_uri_with_error (uri, &error))) {
+      MONGOC_ERROR ("%s", error.message);
+   }
+
+   return client;
+}
+
+
+mongoc_client_t *
+mongoc_client_new_from_uri_with_error (const mongoc_uri_t *uri,
+                                       bson_error_t *error)
+{
+   mongoc_client_t *client;
    mongoc_topology_t *topology;
 
+
+   ENTRY;
+
+   BSON_ASSERT (uri);
+
+#ifndef MONGOC_ENABLE_SSL
+   if (mongoc_uri_get_tls (uri)) {
+      bson_set_error (
+         error,
+         MONGOC_ERROR_COMMAND,
+         MONGOC_ERROR_COMMAND_INVALID_ARG,
+         "Can't create SSL client, SSL not enabled in this build.");
+      RETURN (NULL);
+   }
+#endif
+
    topology = mongoc_topology_new (uri, true);
 
-   /* topology->uri may be different from uri: if this is a mongodb+srv:// URI
-    * then mongoc_topology_new has fetched SRV and TXT records and updated its
-    * uri from them.
-    */
-   return _mongoc_client_new_from_uri (topology);
+   if (!topology->valid) {
+      if (error) {
+         memcpy (error, &topology->scanner->error, sizeof (bson_error_t));
+      }
+
+      mongoc_topology_destroy (topology);
+
+      RETURN (NULL);
+   }
+
+   client = _mongoc_client_new_from_topology (topology);
+   BSON_ASSERT (client);
+
+   RETURN (client);
 }
 
-/*
- *--------------------------------------------------------------------------
- *
- * _mongoc_client_new_from_uri --
- *
- *       Create a new mongoc_client_t for a given topology object.
- *
- * Returns:
- *       A newly allocated mongoc_client_t.
- *
- * Side effects:
- *       None.
- *
- *--------------------------------------------------------------------------
- */
 
+/* precondition: topology is valid */
 mongoc_client_t *
-_mongoc_client_new_from_uri (mongoc_topology_t *topology)
+_mongoc_client_new_from_topology (mongoc_topology_t *topology)
 {
    mongoc_client_t *client;
    const mongoc_read_prefs_t *read_prefs;
@@ -1120,13 +1130,7 @@ _mongoc_client_new_from_uri (mongoc_topology_t *topology)
    const char *appname;
 
    BSON_ASSERT (topology);
-
-#ifndef MONGOC_ENABLE_SSL
-   if (mongoc_uri_get_tls (topology->uri)) {
-      MONGOC_ERROR ("Can't create SSL client, SSL not enabled in this build.");
-      return NULL;
-   }
-#endif
+   BSON_ASSERT (topology->valid);
 
    client = (mongoc_client_t *) bson_malloc0 (sizeof *client);
    client->uri = mongoc_uri_copy (topology->uri);
@@ -1206,6 +1210,7 @@ mongoc_client_destroy (mongoc_client_t *client)
       mongoc_cluster_destroy (&client->cluster);
       mongoc_uri_destroy (client->uri);
       mongoc_set_destroy (client->client_sessions);
+      mongoc_server_api_destroy (client->api);
 
 #ifdef MONGOC_ENABLE_SSL
       _mongoc_ssl_opts_cleanup (&client->ssl_opts, true);
@@ -1283,6 +1288,17 @@ mongoc_client_start_session (mongoc_client_t *client,
       csid = (uint32_t) _mongoc_rand_simple (&client->csid_rand_seed);
    } while (mongoc_set_get (client->client_sessions, csid));
 
+   /* causal consistency and snapshot cannot both be set. */
+   if (opts && mongoc_session_opts_get_causal_consistency (opts) &&
+       mongoc_session_opts_get_snapshot (opts)) {
+      bson_set_error (
+         error,
+         MONGOC_ERROR_CLIENT,
+         MONGOC_ERROR_CLIENT_SESSION_FAILURE,
+         "Only one of causal consistency and snapshot can be enabled.");
+      _mongoc_client_push_server_session (client, ss);
+      RETURN (NULL);
+   }
    cs = _mongoc_client_session_new (client, ss, opts, csid);
 
    /* remember session so if we see its client_session_id in a command, we can
@@ -1709,9 +1725,8 @@ retry:
       retry_server_stream = mongoc_cluster_stream_for_writes (
          &client->cluster, parts->assembled.session, NULL, &ignored_error);
 
-      if (retry_server_stream &&
-          retry_server_stream->sd->max_wire_version >=
-             WIRE_VERSION_RETRY_WRITES) {
+      if (retry_server_stream && retry_server_stream->sd->max_wire_version >=
+                                    WIRE_VERSION_RETRY_WRITES) {
          parts->assembled.server_stream = retry_server_stream;
          bson_destroy (reply);
          GOTO (retry);
@@ -1760,9 +1775,8 @@ retry:
     * a new readable stream and retry. If server selection fails or the selected
     * server does not support retryable reads, fall through and allow the
     * original error to be reported. */
-   if (is_retryable &&
-       _mongoc_read_error_get_type (ret, error, reply) ==
-          MONGOC_READ_ERR_RETRY) {
+   if (is_retryable && _mongoc_read_error_get_type (ret, error, reply) ==
+                          MONGOC_READ_ERR_RETRY) {
       bson_error_t ignored_error;
 
       /* each read command may be retried at most once */
@@ -1777,11 +1791,11 @@ retry:
                                           parts->read_prefs,
                                           parts->assembled.session,
                                           NULL,
+                                          /* Not aggregate-with-write */ false,
                                           &ignored_error);
 
-      if (retry_server_stream &&
-          retry_server_stream->sd->max_wire_version >=
-             WIRE_VERSION_RETRY_READS) {
+      if (retry_server_stream && retry_server_stream->sd->max_wire_version >=
+                                    WIRE_VERSION_RETRY_READS) {
          parts->assembled.server_stream = retry_server_stream;
          bson_destroy (reply);
          GOTO (retry);
@@ -1866,7 +1880,12 @@ mongoc_client_command_simple (mongoc_client_t *client,
     * preference argument."
     */
    server_stream =
-      mongoc_cluster_stream_for_reads (cluster, read_prefs, NULL, reply, error);
+      mongoc_cluster_stream_for_reads (cluster,
+                                       read_prefs,
+                                       NULL,
+                                       reply,
+                                       /* Not aggregate-with-write */ false,
+                                       error);
 
    if (server_stream) {
       ret = _mongoc_client_command_with_stream (
@@ -2019,14 +2038,19 @@ _mongoc_client_command_with_opts (mongoc_client_t *client,
                                            error);
 
       if (server_stream && server_stream->sd->type != MONGOC_SERVER_MONGOS) {
-         parts.user_query_flags |= MONGOC_QUERY_SLAVE_OK;
+         parts.user_query_flags |= MONGOC_QUERY_SECONDARY_OK;
       }
    } else if (parts.is_write_command) {
       server_stream =
          mongoc_cluster_stream_for_writes (cluster, cs, reply_ptr, error);
    } else {
       server_stream =
-         mongoc_cluster_stream_for_reads (cluster, prefs, cs, reply_ptr, error);
+         mongoc_cluster_stream_for_reads (cluster,
+                                          prefs,
+                                          cs,
+                                          reply_ptr,
+                                          /* Not aggregate-with-write */ false,
+                                          error);
    }
 
    if (!server_stream) {
@@ -2335,6 +2359,9 @@ _mongoc_client_monitor_op_killcursors (mongoc_cluster_t *cluster,
                                     operation_id,
                                     &server_stream->sd->host,
                                     server_stream->sd->id,
+                                    &server_stream->sd->service_id,
+                                    server_stream->sd->server_connection_id,
+                                    NULL,
                                     client->apm_context);
 
    client->apm_callbacks.started (&event);
@@ -2381,6 +2408,9 @@ _mongoc_client_monitor_op_killcursors_succeeded (
                                       operation_id,
                                       &server_stream->sd->host,
                                       server_stream->sd->id,
+                                      &server_stream->sd->service_id,
+                                      server_stream->sd->server_connection_id,
+                                      false,
                                       client->apm_context);
 
    client->apm_callbacks.succeeded (&event);
@@ -2423,6 +2453,9 @@ _mongoc_client_monitor_op_killcursors_failed (
                                    operation_id,
                                    &server_stream->sd->host,
                                    server_stream->sd->id,
+                                   &server_stream->sd->service_id,
+                                   server_stream->sd->server_connection_id,
+                                   false,
                                    client->apm_context);
 
    client->apm_callbacks.failed (&event);
@@ -2503,7 +2536,7 @@ _mongoc_client_killcursors_command (mongoc_cluster_t *cluster,
 
    _mongoc_client_prepare_killcursors_command (cursor_id, collection, &command);
    mongoc_cmd_parts_init (
-      &parts, cluster->client, db, MONGOC_QUERY_SLAVE_OK, &command);
+      &parts, cluster->client, db, MONGOC_QUERY_SECONDARY_OK, &command);
    parts.assembled.operation_id = ++cluster->operation_id;
    mongoc_cmd_parts_set_session (&parts, cs);
 
@@ -2549,36 +2582,35 @@ _mongoc_client_killcursors_command (mongoc_cluster_t *cluster,
 void
 mongoc_client_kill_cursor (mongoc_client_t *client, int64_t cursor_id)
 {
-   mongoc_topology_t *topology;
-   mongoc_server_description_t *selected_server;
+   mongoc_topology_t *const topology =
+      BSON_ASSERT_PTR_INLINE (client)->topology;
+   mongoc_server_description_t const *selected_server;
    mongoc_read_prefs_t *read_prefs;
    bson_error_t error;
    uint32_t server_id = 0;
+   mc_shared_tpld td = mc_tpld_take_ref (topology);
 
-   topology = client->topology;
    read_prefs = mongoc_read_prefs_new (MONGOC_READ_PRIMARY);
 
-   bson_mutex_lock (&topology->mutex);
-   if (!mongoc_topology_compatible (&topology->description, NULL, &error)) {
+   if (!mongoc_topology_compatible (td.ptr, NULL, &error)) {
       MONGOC_ERROR ("Could not kill cursor: %s", error.message);
-      bson_mutex_unlock (&topology->mutex);
+      mc_tpld_drop_ref (&td);
       mongoc_read_prefs_destroy (read_prefs);
       return;
    }
 
    /* see if there's a known writable server - do no I/O or retries */
    selected_server =
-      mongoc_topology_description_select (&topology->description,
+      mongoc_topology_description_select (td.ptr,
                                           MONGOC_SS_WRITE,
                                           read_prefs,
+                                          NULL /* chosen read mode */,
                                           topology->local_threshold_msec);
 
    if (selected_server) {
       server_id = selected_server->id;
    }
 
-   bson_mutex_unlock (&topology->mutex);
-
    if (server_id) {
       _mongoc_client_kill_cursor (client,
                                   server_id,
@@ -2592,6 +2624,7 @@ mongoc_client_kill_cursor (mongoc_client_t *client, int64_t cursor_id)
    }
 
    mongoc_read_prefs_destroy (read_prefs);
+   mc_tpld_drop_ref (&td);
 }
 
 
@@ -2744,7 +2777,13 @@ _mongoc_client_set_apm_callbacks_private (mongoc_client_t *client,
 
    /* A client pool sets APM callbacks for the entire pool. */
    if (client->topology->single_threaded) {
-      mongoc_topology_set_apm_callbacks (client->topology, callbacks, context);
+      mongoc_topology_set_apm_callbacks (
+         client->topology,
+         /* We are safe to modify the shared_descr directly, since we are
+          * single-threaded */
+         mc_tpld_unsafe_get_mutable (client->topology),
+         callbacks,
+         context);
    }
 
    return true;
@@ -2765,13 +2804,18 @@ mongoc_client_set_apm_callbacks (mongoc_client_t *client,
    return _mongoc_client_set_apm_callbacks_private (client, callbacks, context);
 }
 
-
 mongoc_server_description_t *
 mongoc_client_get_server_description (mongoc_client_t *client,
                                       uint32_t server_id)
 {
-   /* the error info isn't useful */
-   return mongoc_topology_server_by_id (client->topology, server_id, NULL);
+   mongoc_server_description_t *ret;
+   mc_shared_tpld td = mc_tpld_take_ref (client->topology);
+   mongoc_server_description_t const *sd =
+      mongoc_topology_description_server_by_id_const (
+         td.ptr, server_id, NULL /* <- the error info isn't useful */);
+   ret = mongoc_server_description_new_copy (sd);
+   mc_tpld_drop_ref (&td);
+   return ret;
 }
 
 
@@ -2779,21 +2823,12 @@ mongoc_server_description_t **
 mongoc_client_get_server_descriptions (const mongoc_client_t *client,
                                        size_t *n /* OUT */)
 {
-   mongoc_topology_t *topology;
-   mongoc_server_description_t **sds;
-
-   BSON_ASSERT (client);
-   BSON_ASSERT (n);
-
-   topology = client->topology;
-
-   /* in case the client is pooled */
-   bson_mutex_lock (&topology->mutex);
-
-   sds = mongoc_topology_description_get_servers (&topology->description, n);
-
-   bson_mutex_unlock (&topology->mutex);
-
+   mc_shared_tpld td =
+      mc_tpld_take_ref (BSON_ASSERT_PTR_INLINE (client)->topology);
+   mongoc_server_description_t **const sds =
+      mongoc_topology_description_get_servers (td.ptr,
+                                               BSON_ASSERT_PTR_INLINE (n));
+   mc_tpld_drop_ref (&td);
    return sds;
 }
 
@@ -2833,7 +2868,8 @@ mongoc_client_select_server (mongoc_client_t *client,
       return NULL;
    }
 
-   sd = mongoc_topology_select (client->topology, optype, prefs, error);
+   sd = mongoc_topology_select (
+      client->topology, optype, prefs, NULL /* chosen read mode */, error);
    if (!sd) {
       return NULL;
    }
@@ -2845,7 +2881,8 @@ mongoc_client_select_server (mongoc_client_t *client,
 
    /* check failed, retry once */
    mongoc_server_description_destroy (sd);
-   sd = mongoc_topology_select (client->topology, optype, prefs, error);
+   sd = mongoc_topology_select (
+      client->topology, optype, prefs, NULL /* chosen read mode */, error);
    if (sd) {
       return sd;
    }
@@ -2978,10 +3015,10 @@ _mongoc_client_end_sessions (mongoc_client_t *client)
    mongoc_cluster_t *cluster = &client->cluster;
    bool r;
 
-   if (t->session_pool) {
+   while (!mongoc_server_session_pool_is_empty (t->session_pool)) {
       prefs = mongoc_read_prefs_new (MONGOC_READ_PRIMARY_PREFERRED);
-      server_id =
-         mongoc_topology_select_server_id (t, MONGOC_SS_READ, prefs, &error);
+      server_id = mongoc_topology_select_server_id (
+         t, MONGOC_SS_READ, prefs, NULL /* chosen read mode */, &error);
 
       mongoc_read_prefs_destroy (prefs);
       if (!server_id) {
@@ -3000,7 +3037,7 @@ _mongoc_client_end_sessions (mongoc_client_t *client)
       /* end sessions in chunks */
       while (_mongoc_topology_end_sessions_cmd (t, &cmd)) {
          mongoc_cmd_parts_init (
-            &parts, client, "admin", MONGOC_QUERY_SLAVE_OK, &cmd);
+            &parts, client, "admin", MONGOC_QUERY_SECONDARY_OK, &cmd);
          parts.assembled.operation_id = ++cluster->operation_id;
          parts.prohibit_lsid = true;
 
@@ -3052,7 +3089,7 @@ mongoc_client_reset (mongoc_client_t *client)
    client->client_sessions = mongoc_set_new (8, NULL, NULL);
 
    /* Server sessions are owned by us, so we clear the pool on reset. */
-   _mongoc_topology_clear_session_pool (client->topology);
+   mongoc_server_session_pool_clear (client->topology->session_pool);
 }
 
 mongoc_change_stream_t *
@@ -3078,3 +3115,63 @@ mongoc_client_enable_auto_encryption (mongoc_client_t *client,
    }
    return _mongoc_cse_client_enable_auto_encryption (client, opts, error);
 }
+
+bool
+mongoc_client_set_server_api (mongoc_client_t *client,
+                              const mongoc_server_api_t *api,
+                              bson_error_t *error)
+{
+   BSON_ASSERT_PARAM (client);
+   BSON_ASSERT_PARAM (api);
+
+   if (client->is_pooled) {
+      bson_set_error (
+         error,
+         MONGOC_ERROR_CLIENT,
+         MONGOC_ERROR_CLIENT_API_FROM_POOL,
+         "Cannot set server api on a client checked out from a pool");
+      return false;
+   }
+
+   if (mongoc_client_uses_server_api (client)) {
+      bson_set_error (error,
+                      MONGOC_ERROR_CLIENT,
+                      MONGOC_ERROR_CLIENT_API_ALREADY_SET,
+                      "Cannot set server api more than once per client");
+      return false;
+   }
+
+   client->api = mongoc_server_api_copy (api);
+   _mongoc_topology_scanner_set_server_api (client->topology->scanner, api);
+   return true;
+}
+
+mongoc_server_description_t *
+mongoc_client_get_handshake_description (mongoc_client_t *client,
+                                         uint32_t server_id,
+                                         bson_t *opts,
+                                         bson_error_t *error)
+{
+   mongoc_server_stream_t *server_stream;
+   mongoc_server_description_t *sd;
+
+   server_stream = mongoc_cluster_stream_for_server (&client->cluster,
+                                                     server_id,
+                                                     true /* reconnect */,
+                                                     NULL /* client session */,
+                                                     NULL /* reply */,
+                                                     error);
+   if (!server_stream) {
+      return NULL;
+   }
+
+   sd = mongoc_server_description_new_copy (server_stream->sd);
+   mongoc_server_stream_cleanup (server_stream);
+   return sd;
+}
+
+bool
+mongoc_client_uses_server_api (const mongoc_client_t *client)
+{
+   return mongoc_topology_uses_server_api (client->topology);
+}
