@@ -14,11 +14,13 @@
  * limitations under the License.
  */
 
-#include <bson.h>
+#include <bson/bson.h>
 
 #include "mongoc-client-private.h"
 #include "mongoc-client-session-private.h"
+#include "mongoc-client-side-encryption-private.h"
 #include "mongoc-error.h"
+#include "mongoc-error-private.h"
 #include "mongoc-trace-private.h"
 #include "mongoc-write-command-private.h"
 #include "mongoc-write-command-legacy-private.h"
@@ -118,7 +120,11 @@ _mongoc_write_command_update_append (mongoc_write_command_t *command,
 
    bson_init (&document);
    BSON_APPEND_DOCUMENT (&document, "q", selector);
-   BSON_APPEND_DOCUMENT (&document, "u", update);
+   if (_mongoc_document_is_pipeline (update)) {
+      BSON_APPEND_ARRAY (&document, "u", update);
+   } else {
+      BSON_APPEND_DOCUMENT (&document, "u", update);
+   }
    if (opts) {
       bson_concat (&document, opts);
    }
@@ -194,8 +200,7 @@ _mongoc_write_command_init_insert (mongoc_write_command_t *command, /* IN */
                                    const bson_t *document,          /* IN */
                                    const bson_t *cmd_opts,          /* IN */
                                    mongoc_bulk_write_flags_t flags, /* IN */
-                                   int64_t operation_id,            /* IN */
-                                   bool allow_bulk_op_insert)       /* IN */
+                                   int64_t operation_id)            /* IN */
 {
    ENTRY;
 
@@ -204,7 +209,6 @@ _mongoc_write_command_init_insert (mongoc_write_command_t *command, /* IN */
    _mongoc_write_command_init_bulk (
       command, MONGOC_WRITE_COMMAND_INSERT, flags, operation_id, cmd_opts);
 
-   command->u.insert.allow_bulk_op_insert = (uint8_t) allow_bulk_op_insert;
    /* must handle NULL document from mongoc_collection_insert_bulk */
    if (document) {
       _mongoc_write_command_insert_append (command, document);
@@ -218,8 +222,7 @@ void
 _mongoc_write_command_init_insert_idl (mongoc_write_command_t *command,
                                        const bson_t *document,
                                        const bson_t *cmd_opts,
-                                       int64_t operation_id,
-                                       bool allow_bulk_op_insert)
+                                       int64_t operation_id)
 {
    mongoc_bulk_write_flags_t flags = MONGOC_BULK_WRITE_FLAGS_INIT;
 
@@ -230,7 +233,6 @@ _mongoc_write_command_init_insert_idl (mongoc_write_command_t *command,
    _mongoc_write_command_init_bulk (
       command, MONGOC_WRITE_COMMAND_INSERT, flags, operation_id, cmd_opts);
 
-   command->u.insert.allow_bulk_op_insert = (uint8_t) allow_bulk_op_insert;
    /* must handle NULL document from mongoc_collection_insert_bulk */
    if (document) {
       _mongoc_write_command_insert_append (command, document);
@@ -288,6 +290,7 @@ void
 _mongoc_write_command_init_update (mongoc_write_command_t *command, /* IN */
                                    const bson_t *selector,          /* IN */
                                    const bson_t *update,            /* IN */
+                                   const bson_t *cmd_opts,          /* IN */
                                    const bson_t *opts,              /* IN */
                                    mongoc_bulk_write_flags_t flags, /* IN */
                                    int64_t operation_id)            /* IN */
@@ -299,7 +302,7 @@ _mongoc_write_command_init_update (mongoc_write_command_t *command, /* IN */
    BSON_ASSERT (update);
 
    _mongoc_write_command_init_bulk (
-      command, MONGOC_WRITE_COMMAND_UPDATE, flags, operation_id, NULL);
+      command, MONGOC_WRITE_COMMAND_UPDATE, flags, operation_id, cmd_opts);
    _mongoc_write_command_update_append (command, selector, update, opts);
 
    EXIT;
@@ -310,6 +313,7 @@ void
 _mongoc_write_command_init_update_idl (mongoc_write_command_t *command,
                                        const bson_t *selector,
                                        const bson_t *update,
+                                       const bson_t *cmd_opts,
                                        const bson_t *opts,
                                        int64_t operation_id)
 {
@@ -320,7 +324,7 @@ _mongoc_write_command_init_update_idl (mongoc_write_command_t *command,
    BSON_ASSERT (command);
 
    _mongoc_write_command_init_bulk (
-      command, MONGOC_WRITE_COMMAND_UPDATE, flags, operation_id, NULL);
+      command, MONGOC_WRITE_COMMAND_UPDATE, flags, operation_id, cmd_opts);
    _mongoc_write_command_update_append (command, selector, update, opts);
 
    EXIT;
@@ -342,11 +346,10 @@ _mongoc_write_command_init (bson_t *doc,
    BSON_APPEND_UTF8 (doc, gCommandNames[command->type], collection);
    BSON_APPEND_BOOL (doc, "ordered", command->flags.ordered);
 
-   if (command->flags.bypass_document_validation !=
-       MONGOC_BYPASS_DOCUMENT_VALIDATION_DEFAULT) {
+   if (command->flags.bypass_document_validation) {
       BSON_APPEND_BOOL (doc,
                         "bypassDocumentValidation",
-                        !!command->flags.bypass_document_validation);
+                        command->flags.bypass_document_validation);
    }
 
    EXIT;
@@ -415,7 +418,7 @@ _mongoc_write_command_will_overflow (uint32_t len_so_far,
    /* max BSON object size + 16k bytes.
     * server guarantees there is enough room: SERVER-10643
     */
-   int32_t max_cmd_size = max_bson_size + 16384;
+   int32_t max_cmd_size = max_bson_size + BSON_OBJECT_ALLOWANCE;
 
    BSON_ASSERT (max_bson_size);
 
@@ -466,12 +469,11 @@ _mongoc_write_opmsg (mongoc_write_command_t *command,
    BSON_ASSERT (server_stream);
    BSON_ASSERT (collection);
 
-/* MongoDB has a extra allowance to allow updating 16mb document,
- * as the update operators would otherwise overflow the 16mb object limit
- */
-#define BSON_OBJECT_ALLOWANCE (16 * 1024)
    max_bson_obj_size = mongoc_server_stream_max_bson_obj_size (server_stream);
    max_msg_size = mongoc_server_stream_max_msg_size (server_stream);
+   if (_mongoc_cse_is_enabled (client)) {
+      max_msg_size = MONGOC_REDUCED_MAX_MSG_SIZE_FOR_FLE;
+   }
    max_document_count =
       mongoc_server_stream_max_write_batch_size (server_stream);
 
@@ -540,7 +542,8 @@ _mongoc_write_opmsg (mongoc_write_command_t *command,
          result->failed = true;
          break;
 
-      } else if ((payload_batch_size + header) + len <= max_msg_size) {
+      } else if ((payload_batch_size + header) + len <= max_msg_size ||
+                 document_count == 0) {
          /* The current batch is still under max batch size in bytes */
          payload_batch_size += len;
 
@@ -582,6 +585,11 @@ _mongoc_write_opmsg (mongoc_write_command_t *command,
          ret = mongoc_cluster_run_command_monitored (
             &client->cluster, &parts.assembled, &reply, error);
 
+         if (parts.is_retryable_write) {
+            _mongoc_write_error_handle_labels (
+               ret, error, &reply, server_stream->sd->max_wire_version);
+         }
+
          /* Add this batch size so we skip these documents next time */
          payload_total_offset += payload_batch_size;
          payload_batch_size = 0;
@@ -591,8 +599,11 @@ _mongoc_write_opmsg (mongoc_write_command_t *command,
           * the selected server does not support retryable writes, fall through
           * and allow the original error to be reported. */
          error_type = _mongoc_write_error_get_type (&reply);
-         if (!ret && is_retryable && (error->domain == MONGOC_ERROR_STREAM ||
-                                      error_type == MONGOC_WRITE_ERR_RETRY)) {
+         if (is_retryable) {
+            _mongoc_write_error_update_if_unsupported_storage_engine (
+               ret, error, &reply);
+         }
+         if (is_retryable && error_type == MONGOC_WRITE_ERR_RETRY) {
             bson_error_t ignored_error;
 
             /* each write command may be retried at most once */
@@ -616,7 +627,12 @@ _mongoc_write_opmsg (mongoc_write_command_t *command,
 
          if (!ret) {
             result->failed = true;
-            result->must_stop = true;
+            /* Stop for ordered bulk writes or when the server stream has been
+             * properly invalidated (e.g., due to a network error). */
+            if (command->flags.ordered || !mongoc_cluster_stream_valid (
+                                             &client->cluster, server_stream)) {
+               result->must_stop = true;
+            }
          }
 
          /* Result merge needs to know the absolute index for a document
@@ -629,12 +645,18 @@ _mongoc_write_opmsg (mongoc_write_command_t *command,
          bson_destroy (&reply);
       }
       /* While we have more documents to write */
-   } while (payload_total_offset < command->payload.len);
+   } while (payload_total_offset < command->payload.len && !result->must_stop);
 
    bson_destroy (&cmd);
    mongoc_cmd_parts_cleanup (&parts);
 
    if (retry_server_stream) {
+      if (ret) {
+         /* if a retry succeeded, report that in the result so bulk write can
+          * use the newly selected server. */
+         result->retry_server_id =
+            mongoc_server_description_id (retry_server_stream->sd);
+      }
       mongoc_server_stream_cleanup (retry_server_stream);
    }
 
@@ -677,6 +699,41 @@ _append_array_from_command (mongoc_write_command_t *command, bson_t *bson)
    bson_reader_destroy (reader);
 }
 
+/* Assemble the base @cmd with all of the command options.
+ * @parts is always initialized, even on error.
+ * This is called twice in _mongoc_write_opquery.
+ * Once with no payload documents, to determine the total size. And once with
+ * payload documents, to send the final command. */
+static bool
+_assemble_cmd (bson_t *cmd,
+               mongoc_write_command_t *command,
+               mongoc_client_t *client,
+               mongoc_server_stream_t *server_stream,
+               const char *database,
+               const mongoc_write_concern_t *write_concern,
+               mongoc_cmd_parts_t *parts,
+               bson_error_t *error)
+{
+   bool ret;
+   bson_iter_t iter;
+
+   mongoc_cmd_parts_init (parts, client, database, MONGOC_QUERY_NONE, cmd);
+   parts->is_write_command = true;
+   parts->assembled.operation_id = command->operation_id;
+
+   ret = mongoc_cmd_parts_set_write_concern (
+      parts, write_concern, server_stream->sd->max_wire_version, error);
+   if (ret) {
+      BSON_ASSERT (bson_iter_init (&iter, &command->cmd_opts));
+      ret = mongoc_cmd_parts_append_opts (
+         parts, &iter, server_stream->sd->max_wire_version, error);
+   }
+   if (ret) {
+      ret = mongoc_cmd_parts_assemble (parts, server_stream, error);
+   }
+   return ret;
+}
+
 static void
 _mongoc_write_opquery (mongoc_write_command_t *command,
                        mongoc_client_t *client,
@@ -689,12 +746,10 @@ _mongoc_write_opquery (mongoc_write_command_t *command,
                        bson_error_t *error)
 {
    mongoc_cmd_parts_t parts;
-   bson_iter_t iter;
    const char *key;
    uint32_t len = 0;
    bson_t ar;
    bson_t cmd;
-   bson_t reply;
    char str[16];
    bool has_more;
    bool ret = false;
@@ -726,10 +781,30 @@ again:
    i = 0;
 
    _mongoc_write_command_init (&cmd, command, collection);
+   /* If any part of assembling failed, return with failure. */
+   if (!_assemble_cmd (&cmd,
+                       command,
+                       client,
+                       server_stream,
+                       database,
+                       write_concern,
+                       &parts,
+                       error)) {
+      result->failed = true;
+      bson_destroy (&cmd);
+      mongoc_cmd_parts_cleanup (&parts);
+      EXIT;
+   }
 
-   /* 1 byte to specify array type, 1 byte for field name's null terminator */
-   overhead = cmd.len + 2 + gCommandFieldLens[command->type];
-
+   /* Use the assembled command to compute the overhead, since it may be a new
+    * BSON document with options applied. If no options were applied, then
+    * parts.assembled.command points to cmd. The constant 2 is due to 1 byte to
+    * specify array type and 1 byte for field name's null terminator. */
+   overhead =
+      parts.assembled.command->len + 2 + gCommandFieldLens[command->type];
+   /* Toss out the assembled command, we'll assemble again after adding all of
+    * the payload documents. */
+   mongoc_cmd_parts_cleanup (&parts);
 
    reader = bson_reader_new_from_data (command->payload.data + data_offset,
                                        command->payload.len - data_offset);
@@ -767,42 +842,27 @@ again:
          data_offset += len;
       }
    } else {
-      mongoc_cmd_parts_init (&parts, client, database, MONGOC_QUERY_NONE, &cmd);
-      parts.is_write_command = true;
-      parts.assembled.operation_id = command->operation_id;
-      if (!mongoc_cmd_parts_set_write_concern (
-             &parts,
-             write_concern,
-             server_stream->sd->max_wire_version,
-             error)) {
-         bson_reader_destroy (reader);
-         bson_destroy (&cmd);
-         mongoc_cmd_parts_cleanup (&parts);
-         EXIT;
-      }
+      bson_t reply;
 
-      BSON_ASSERT (bson_iter_init (&iter, &command->cmd_opts));
-      if (!mongoc_cmd_parts_append_opts (
-             &parts, &iter, server_stream->sd->max_wire_version, error)) {
-         bson_reader_destroy (reader);
-         bson_destroy (&cmd);
-         mongoc_cmd_parts_cleanup (&parts);
-         EXIT;
-      }
-
-      ret = mongoc_cmd_parts_assemble (&parts, server_stream, error);
+      ret = _assemble_cmd (&cmd,
+                           command,
+                           client,
+                           server_stream,
+                           database,
+                           write_concern,
+                           &parts,
+                           error);
       if (ret) {
          ret = mongoc_cluster_run_command_monitored (
             &client->cluster, &parts.assembled, &reply, error);
       } else {
-         /* assembling failed */
-         result->must_stop = true;
          bson_init (&reply);
       }
 
       if (!ret) {
          result->failed = true;
-         if (bson_empty (&reply)) {
+         if (bson_empty (&reply) ||
+             !mongoc_cluster_stream_valid (&client->cluster, server_stream)) {
             /* assembling failed, or a network error running the command */
             result->must_stop = true;
          }
@@ -934,8 +994,37 @@ _mongoc_write_command_execute_idl (mongoc_write_command_t *command,
       }
    }
 
-   if (command->flags.bypass_document_validation !=
-       MONGOC_BYPASS_DOCUMENT_VALIDATION_DEFAULT) {
+   if (command->flags.has_update_hint) {
+      if (server_stream->sd->max_wire_version <
+             WIRE_VERSION_HINT_SERVER_SIDE_ERROR ||
+          (server_stream->sd->max_wire_version < WIRE_VERSION_UPDATE_HINT &&
+           !mongoc_write_concern_is_acknowledged (crud->writeConcern))) {
+         bson_set_error (
+            &result->error,
+            MONGOC_ERROR_COMMAND,
+            MONGOC_ERROR_PROTOCOL_BAD_WIRE_VERSION,
+            "The selected server does not support hint for update");
+         result->failed = true;
+         EXIT;
+      }
+   }
+
+   if (command->flags.has_delete_hint) {
+      if (server_stream->sd->max_wire_version <
+             WIRE_VERSION_HINT_SERVER_SIDE_ERROR ||
+          (server_stream->sd->max_wire_version < WIRE_VERSION_DELETE_HINT &&
+           !mongoc_write_concern_is_acknowledged (crud->writeConcern))) {
+         bson_set_error (
+            &result->error,
+            MONGOC_ERROR_COMMAND,
+            MONGOC_ERROR_COMMAND_INVALID_ARG,
+            "The selected server does not support hint for delete");
+         result->failed = true;
+         EXIT;
+      }
+   }
+
+   if (command->flags.bypass_document_validation) {
       if (!mongoc_write_concern_is_acknowledged (crud->writeConcern)) {
          result->failed = true;
          bson_set_error (
@@ -1222,7 +1311,7 @@ _mongoc_write_result_merge (mongoc_write_result_t *result,   /* IN */
       /* writeConcernError is a subdocument in the server response
        * append it to the result->writeConcernErrors array */
       bson_iter_document (&iter, &len, &data);
-      bson_init_static (&write_concern_error, data, len);
+      BSON_ASSERT (bson_init_static (&write_concern_error, data, len));
 
       bson_uint32_to_string (
          result->n_writeConcernErrors, &key, str, sizeof str);
@@ -1279,7 +1368,7 @@ _set_error_from_response (bson_t *bson_array,
             while (bson_iter_next (&doc_iter)) {
                /* use the first error code we find */
                if (BSON_ITER_IS_KEY (&doc_iter, "code") && code == 0) {
-                  code = bson_iter_int32 (&doc_iter);
+                  code = (uint32_t) bson_iter_as_int64 (&doc_iter);
                } else if (BSON_ITER_IS_KEY (&doc_iter, "errmsg")) {
                   errmsg = bson_iter_utf8 (&doc_iter, NULL);
 
@@ -1432,35 +1521,81 @@ _mongoc_write_result_complete (
 }
 
 
+/*--------------------------------------------------------------------------
+ *
+ * _mongoc_write_error_get_type --
+ *
+ *       Checks if the error or reply from a write command is considered
+ *       retryable according to the retryable writes spec. Checks both
+ *       for a client error (a network exception) and a server error in
+ *       the reply. @cmd_ret and @cmd_err come from the result of a
+ *       write_command function. This function should be called after
+ *       error labels are appended in _mongoc_write_error_handle_labels,
+ *       which should be called after mongoc_cluster_run_command_monitored.
+ *
+ *
+ * Return:
+ *       A mongoc_write_error_type_t indicating the type of error (if any).
+ *
+ *--------------------------------------------------------------------------
+ */
 mongoc_write_err_type_t
-_mongoc_write_error_get_type (const bson_t *reply)
+_mongoc_write_error_get_type (bson_t *reply)
 {
    bson_error_t error;
+
+   if (mongoc_error_has_label (reply, RETRYABLE_WRITE_ERROR)) {
+      return MONGOC_WRITE_ERR_RETRY;
+   }
+
+   /* check for a server error. */
    if (_mongoc_cmd_check_ok_no_wce (
           reply, MONGOC_ERROR_API_VERSION_2, &error)) {
       return MONGOC_WRITE_ERR_NONE;
    }
 
    switch (error.code) {
-   case 11600: /* InterruptedAtShutdown */
-   case 11602: /* InterruptedDueToReplStateChange */
-   case 10107: /* NotMaster */
-   case 13435: /* NotMasterNoSlaveOk */
-   case 13436: /* NotMasterOrSecondary */
-   case 189:   /* PrimarySteppedDown */
-   case 91:    /* ShutdownInProgress */
-   case 7:     /* HostNotFound */
-   case 6:     /* HostUnreachable */
-   case 89:    /* NetworkTimeout */
-   case 9001:  /* SocketException */
-      return MONGOC_WRITE_ERR_RETRY;
    case 64: /* WriteConcernFailed */
       return MONGOC_WRITE_ERR_WRITE_CONCERN;
    default:
-      if (strstr (error.message, "not master") ||
-          strstr (error.message, "node is recovering")) {
-         return MONGOC_WRITE_ERR_RETRY;
-      }
       return MONGOC_WRITE_ERR_OTHER;
    }
 }
+
+/* Returns true and modifies reply and cmd_err. */
+bool
+_mongoc_write_error_update_if_unsupported_storage_engine (bool cmd_ret,
+                                                          bson_error_t *cmd_err,
+                                                          bson_t *reply)
+{
+   bson_error_t server_error;
+
+   if (cmd_ret) {
+      return false;
+   }
+
+   if (_mongoc_cmd_check_ok_no_wce (
+          reply, MONGOC_ERROR_API_VERSION_2, &server_error)) {
+      return false;
+   }
+
+   if (server_error.code == 20 &&
+       strstr (server_error.message, "Transaction numbers") ==
+          server_error.message) {
+      const char *replacement = "This MongoDB deployment does not support "
+                                "retryable writes. Please add "
+                                "retryWrites=false to your connection string.";
+
+      strcpy (cmd_err->message, replacement);
+
+      if (reply) {
+         bson_t *new_reply = bson_new ();
+         bson_copy_to_excluding_noinit (reply, new_reply, "errmsg", NULL);
+         BSON_APPEND_UTF8 (new_reply, "errmsg", replacement);
+         bson_destroy (reply);
+         bson_steal (reply, new_reply);
+      }
+      return true;
+   }
+   return false;
+}
